define(["jquery", "underscore", "backbone", "tpls", "context", "grid", "models/table/schema",
		"events", "models/table/column"
	],
	function($, _, Backbone, Templates, Context, Grid, Schema, Events, Column) {
		return Backbone.View.extend({
			template: Templates["builder/structure"],
			gridSettings: {
				tableClass: "table table-bordered table-condensed",
				sortable: false,
				emptyRow: true
			},
			table: null,
			columns: ["Field", "Type", "CharacterSet", "Null", "Default", "Extra", "Comment"],
			gridColumns: function() {
				this.charsets.unshift("");
				var charsets = {
					values: this.charsets
				};
				var nulls = {
					values: ["YES", "NO"]
				};
				return _.map(this.columns, function(col) {
					if (col === "CharacterSet") {
						return {
							name: col,
							type: "string",
							editor: "SelectEditor",
							editorProps: charsets
						};
					} else if (col === "Null") {
						return {
							name: col,
							type: "string",
							editor: "SelectEditor",
							editorProps: nulls
						};
					} else {
						return {
							name: col,
							type: "string",
							editor: "BasicEditor"
						};
					}
				});
			},
			initialize: function() {
				_.bindAll(this, "updateColumn", "createColumn", "renderGrid", "error", "changeError", "changeSuccess",
					"setTable");

				// listen to global events
				this.listenTo(Events, "table:change", this.setTable);
				this.listenTo(Events, "db:change", this.resetView);
			},
			setTable: function(table) {
				this.fetchSchema();
			},
			resetView: function() {
				this.$el.html(this.template());
			},
			fetchSchema: function() {
				if (Context.table !== this.table) {
					this.$(".alert").remove();
					var schema = new Schema();
					schema.fetch({
						success: this.renderGrid,
						error: this.error
					});
					this.table = Context.table;
				}
			},
			render: function() {

				Events.trigger("builder:render");

				this.$el.html(this.template());

				this.fetchSchema();

				return this;
			},
			/**
			 * Handle schema fetch errors
			 * @param model
			 * @param error
			 */
			error: function(model, error) {
				alert("Failed to fetch table schema.");
			},
			/**
			 * Render sensei grid
			 * @param model
			 * @param data
			 */
			renderGrid: function(model, data) {

				// store initial data for later use
				this.initialData = data.meta;
				this.charsets = data.charsets;
				this.columns = data.columns;
				this.indexes = data.indexes;
				this.foreignKeys = data.foreignKeys;

				// get sensei grid container element
				var $el = this.$(".sensei-grid-builder");

				// grid initialization
				this.grid = $el.grid(
					this.initialData,
					this.gridColumns(),
					this.gridSettings
				);

				// register basic editor
				this.grid.registerEditor(BasicEditor);
				this.grid.registerEditor(SelectEditor);

				// listen to events
				this.grid.events.on("editor:save", this.updateColumn);
				this.grid.events.on("row:save", this.createColumn);
				this.grid.events.on("row:remove", this.deleteColumn);

				// render grid
				this.grid.render();

				// render index grid
				this.renderIndexGrid();

				// render index grid
				this.renderForeignKeyGrid();
			},
			renderIndexGrid: function(model, data) {

				// get sensei grid container element
				var $el = this.$(".sensei-grid-index");

				// grid initialization
				this.indexgrid = $el.grid(
					this.indexes, [{
						name: "Name",
						type: "string"
					}, {
						name: "Type",
						type: "string"
					}, {
						name: "Columns",
						type: "string"
					}, {
						name: "Sub_Part",
						type: "string"
					}, {
						name: "Non_Unique",
						type: "string",
						editor: "BooleanEditor"
					}],
					this.gridSettings);

				// register basic editor
				this.indexgrid.registerEditor(BasicEditor);
				this.indexgrid.registerEditor(SelectEditor);
				this.indexgrid.registerEditor(BooleanEditor);

				this.indexgrid.events.on("editor:save", function() {
					console.log("indexgrid editor:save")
				});

				// render grid
				this.indexgrid.render();
			},
			renderForeignKeyGrid: function(model, data) {

				// get sensei grid container element
				var $el = this.$(".sensei-grid-fk");

				/**
        **Example data**
        Column: "author_id"
        Constraint: "author_id_refs_id_8a197f2f"
        Referenced_Column: "id"
        Referenced_Table: "auth_user"
        Table: "vellum_post"
        */

				// grid initialization
				this.fkgrid = $el.grid(
					this.foreignKeys, [{
						name: "Column",
						type: "string"
					}, {
						name: "Constraint",
						type: "string"
					}, {
						name: "Referenced_Column",
						type: "string"
					}, {
						name: "Referenced_Table",
						type: "string"
					}, {
						name: "Table",
						type: "string"
					}],
					this.gridSettings);

				this.fkgrid.events.on("editor:save", function() {
					console.log("fk editor:save")
				});

				// register basic editor
				this.fkgrid.registerEditor(BasicEditor);
				this.fkgrid.registerEditor(SelectEditor);
				this.fkgrid.registerEditor(BooleanEditor);

				// render grid
				this.fkgrid.render();
			},
			/**
			 * Update an existing column
			 * @param data
			 * @param $cell
			 */
			updateColumn: function(data, $cell) {

				if (!this.grid.getCellStatus($cell)) {
					console.log("nope updateColumn");
					return;
				}

				console.log("updateColumn", data, $cell);
				_.forIn(data, function(value, param) {
					var $row = this.grid.getCellRow($cell);
					var index = $row.index();

					if (this.initialData[index] && this.initialData[index]["Field"]) {

						var column = this.initialData[index]["Field"];
						var data = {
							param: param,
							value: value,
							column: column
						};
						var schema = new Schema();

						schema.save(data, {
							success: this.changeSuccess,
							error: this.changeError
						});
					} else {
						this.changeError();
					}
				}, this);
			},
			/**
			 * Create a new column
			 * @param data
			 * @param $row
			 */
			createColumn: function(data, $row) {
				console.log("create column", data, $row);
				// a little workaround for backbone
				var schema = new Schema();
				schema.setCreateURL();
				var grid = this.grid;
				schema.save(data, {
					success: function(model, response, options) {
						grid.setRowSaved($row);
						Events.trigger("schema:changed");
					},
					error: this.changeError
				});
			},
			/**
			 * Delete a column
			 * @param data
			 * @param row
			 * @param $row
			 */
			deleteColumn: function(data, row, $row) {
				console.log("StructureView.deleteColumn", data, $row);

				var col = new Column({
					title: data["Field"]
				});
				col.destroy({
					success: function() {
						console.log("column dropped", arguments, col);
						Events.trigger("column:dropped", col);
						Events.trigger("schema:changed");
					},
					error: function() {
						alert("Unexpected error: failed to drop column.");
						console.log("StructureView.deleteColumn Error", arguments);
					}
				});
			},
			changeError: function(model, response, options) {
				if (response["responseJSON"] && response["responseJSON"]["error"]) {
					console.log(response);
					alert("Schema update failed. " + response["responseJSON"]["error"]);
				} else {
					alert("Schema update failed: unexpected error.");
				}
			},
			changeSuccess: function(model, response, options) {
				// do nothing for now
				Events.trigger("schema:changed");
			}
		});
	});
